#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "mpi.h"

#include "u.h"

#define SMALL  200
#define ITMAX  100000

int init_mpi(int argc,char *argv[], int* rang, int* nb_process);

void main_loop(params p, double** f, double** u, double** u_new, double** u_exact);


int main(int argc,char *argv[])
{

  /* Variables MPI */
  int rank;          /* Rang du processus MPI */
  int nb_process;    /* Nombre de processus MPI */

  /* Parametres de la simulation definis dans u.h */
  int nx;
  int ny;
  int itmax;
  params p;

  /* Tableaux associes aux fonctions f et u */
  double** f;
  double** u;
  double** u_new;
  double** u_exact;

  double hx, hy;
  char filename[64];

  int i;
  double t0, t1;


  /* Nombre de points interieurs */
  nx = 16;
  ny = nx;

  hx = 1./( (double) nx + 1 );
  hy = 1./( (double) ny + 1);

  /* Nombre max d'iterations */
  if (nx < SMALL) 
    itmax = ITMAX;
  else
    itmax = 10;

  init_mpi(argc, argv, &rank, &nb_process);

  if (rank == 0)
    printf("Nb d'iterations max : %d\n", itmax);

  /* Ajout des lignes fantomes + decoupage MPI */
  ny += 2;
  nx = nx/nb_process + 2;

  /* Initialisation des Parametres */
  p = init_params(itmax, nx, ny, hx, hy, rank, nb_process);
  f = NULL;
  u = NULL;
  u_new = NULL;
  u_exact = NULL;
  init_grids(p, &f, &u, &u_new, &u_exact);

  t0 = MPI_Wtime();

  /* Boucle principale */
  main_loop(p, f, u, u_new, u_exact);

  t1 = MPI_Wtime();

  if (rank == 0) {
    printf("Temps : %f\n", t1-t0);
  }


  /* Fabrication de fichiers pour le plot si pour 16 points interieurs */
  if (ny == 18) {
    if (rank == 0) {
      for (i=0 ; i<16 ; i++) {
        sprintf(filename, "data/u_%d.dat", i);
        remove(filename);
        FILE* file;
        file = fopen(filename,"w");
        fclose(file);
      }
    }
    MPI_Barrier(MPI_COMM_WORLD);
    sprintf(filename, "data/u_%d.dat", rank);
    write_grid(filename,p,u_new);
  }

  /* Finalisation */
  deinit_grids(f, u, u_new, u_exact);
  MPI_Finalize();
  return 0;
}





/* Jacobi Iterations */
void main_loop(params p, double** f, double** u, double** u_new, double** u_exact)
{
  int it;
  int next, prev;
  double **u_tmp;
  double local_error;
  double global_error;
  int tag = 100;
  global_error = 1.;

  next = p.rank+1;
  prev = p.rank-1;
  if (next == p.size) next = MPI_PROC_NULL;
  if (prev == -1    ) prev = MPI_PROC_NULL;

  for (it=1 ; it <= p.itmax ; it++)
  {
    /* Double buffering : on echange les pointeurs de u_new et u */
    u_tmp = u;
    u = u_new;
    u_new = u_tmp;

    /* Echange de donnees avec les voisins  */
    MPI_Sendrecv(u[p.nx-2], p.ny, MPI_DOUBLE, next, tag,
                 u[p.nx-1], p.ny, MPI_DOUBLE, next, tag,
                 MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    MPI_Sendrecv(u[1], p.ny, MPI_DOUBLE, prev, tag,
                 u[0], p.ny, MPI_DOUBLE, prev, tag,
                 MPI_COMM_WORLD, MPI_STATUS_IGNORE);


    /* Calcul de u_new a partir de u et f */
    update_u(p, u, f, u_new);

    /* Calcul de l'erreur dans le sous-domaine */
    local_error = error(p, u_new, u_exact);
    MPI_Allreduce(&local_error, &global_error, 1, MPI_DOUBLE,
      MPI_MAX, MPI_COMM_WORLD);

    if (p.rank == 0)
      if ( it % 100 == 0)  {
          printf("Iteration: %8d     Erreur : %16.10e\n", it, global_error);
    }

    /* Test convergence */
    if (global_error < 1.e-10) break;

  }
}






/* Initialization MPI */
int init_mpi(int argc,char *argv[], int* rank, int* nb_process)
{
  int rc;

  /* Initialisation de l'environnement MPI */
  rc = MPI_Init(&argc,&argv);
  if (rc)
  {
    fprintf(stderr,"Error in MPI_init");
    exit(255);
  }

  /* Qui suis-je? */
  rc = MPI_Comm_rank(MPI_COMM_WORLD,rank);
  if (rc)
  {
    fprintf(stderr,"Error in MPI_Comm_rank");
    exit(255);
  }

  /* Nombre de processus */
  MPI_Comm_size(MPI_COMM_WORLD,nb_process);
  if (rc)
  {
    fprintf(stderr,"Error in MPI_Comm_size");
    exit(255);
  }
  return 0;
}





