#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "u.h"


params init_params(int itmax, int nx, int ny, double hx, double hy, int rank, int size)
{
  params p;

  assert (size > 0);
  assert (rank >= 0);
  assert (nx > 0);
  assert (ny > 0);

  p.rank = rank;
  p.size = size;
  p.nx = nx;
  p.ny = ny;
  p.hx = hx;
  p.hy = hy;
  p.c0 = (0.5*hx*hx*hy*hy)/(hx*hx+hy*hy);
  p.c1 = 1./(hx*hx);
  p.c2 = 1./(hy*hy);

  p.itmax = itmax;

  assert (p.nx > 0);
  assert (p.ny > 0);
  return p;
}


void init_grids(params p, double*** f, double*** u, double*** u_new, double*** u_exact)
{
  int i,j,i0;
  double x,y;
  double hx,  hy;

  *u          = (double**) malloc (p.nx * sizeof(double*) );
  *f          = (double**) malloc (p.nx * sizeof(double*) );
  *u_new      = (double**) malloc (p.nx * sizeof(double*) );
  *u_exact    = (double**) malloc (p.nx * sizeof(double*) );

  (*u)[0]       = (double*)  malloc (p.nx*p.ny * sizeof(double) );
  (*f)[0]       = (double*)  malloc (p.nx*p.ny * sizeof(double) );
  (*u_new)[0]   = (double*)  malloc (p.nx*p.ny * sizeof(double) );
  (*u_exact)[0] = (double*)  malloc (p.nx*p.ny * sizeof(double) );

  for (i=1 ; i<p.nx ; i++) {
    (*u)[i] = (*u)[i-1] + p.ny;
    (*f)[i] = (*f)[i-1] + p.ny;
    (*u_new)[i] = (*u_new)[i-1] + p.ny;
    (*u_exact)[i] = (*u_exact)[i-1] + p.ny;
  }

  for (i=0 ; i<p.nx ; i++) {
    for (j=0 ; j<p.ny ; j++) {
    (*u)[i][j] = 0.;
    (*f)[i][j] = 0.;
    (*u_new)[i][j] = 0.;
    (*u_exact)[i][j] = 0.;
    }
  }

  /* Remplissage des tableaux */
  hx = p.hx;
  hy = p.hy;

  i0 = p.rank * (p.nx-2);
  for (i=0 ; i<p.nx ; i++)
  {
    x = (i0+i)*hx;
    for (j=0 ; j<p.ny ; j++)
    {
      y = j*hy;
      (*u)      [i][j] = 0.;
      (*u_new)  [i][j] = 0.;
      (*f)      [i][j] = 2.*(x*x-x+y*y-y);
      (*u_exact)[i][j] = x*y*(x-1.)*(y-1.);
    }
  }

}

void deinit_grids(double** f, double** u, double** u_new, double** u_exact)
{
  free(f[0]);
  free(u[0]);
  free(u_new[0]);
  free(u_exact[0]);
  free(f);
  free(u);
  free(u_new);
  free(u_exact);
}

/* Calcul de u_new a partir de u */
void update_u(params p, double** u, double** f, double** u_new)
{
  int i,j;
  for (i=1 ; i<p.nx-1 ; i++) {
    for (j=1 ; j<p.ny-1 ; j++) {
      u_new[i][j] = p.c0 * ( p.c1 * ( u[i+1][j] + u[i-1][j] )
        + p.c2 * ( u[i][j+1] + u[i][j-1] ) - f[i][j] );
    }
  }

}


/* Impression */
void write_grid(char* filename, params p, double** f)
{
  int i,j;
  FILE* file;

  file = fopen(filename,"w");
  for (i=1 ; i<p.nx-1 ; i++) {
    for (j=1 ; j<p.ny-1 ; j++) {
      fprintf(file,"%d %d %16.10e\n", i + p.rank*(p.nx-2), j, f[i][j]);
    }
    fprintf(file,"\n");
  }
  fclose(file);
}



/* Calcul de l'erreur */
double error(params p, double** u, double** u_ref)
{
  int i,j;
  double diff;
  double result;
  result = 0.;
  for (i=1 ; i<p.nx-1 ; i++) {
    for (j=1 ; j<p.ny-1 ; j++) {
      diff = fabs(u[i][j] - u_ref[i][j]);
      if (diff > result)
        result = diff;
    }
  }
  return result;
}


